
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>

#include <botnet.h>

#include "cbot.h"

#define debug( fmt, arg... )	\
	do{ if( opt_debug ) printf( fmt, ##arg ); }while(0)

#define BLACK	"irc-black"
#define GREEN	"irc-green"
#define BROWN	"irc-brown"
#define NAVY	"irc-navy"
#define BRICK	"irc-brick"
#define RED	"irc-red"

int opt_debug;
int opt_daemon;
int opt_port = 6667;
char *opt_channel = "#default";
char *opt_nick = "poors";
char *opt_user = "username";
char *opt_realname = "realname";
char *opt_server = "localhost";
char *opt_dirlog = ".";
char *opt_diraddon = ".";

void replace( char *src, char *dest, int ch, char *rep )
{
	int rep_len = strlen( rep );

	do
	{
		if( *src == ch )
		{
			memcpy( dest, rep, rep_len );
			dest += rep_len;
		}
		else
		{
			*dest = *src;
			dest ++;
		}
		src ++;
	}
	while( *src );

	*dest = 0;
}

void logging( const char *class, const char *fmt, ... )
{
	va_list ap;
	time_t now;
	struct tm *ptm;
	char *buf1;
	char *buf2;
	int allocated = 0;
	int len;
	FILE *out;
	char filename[strlen(opt_dirlog) + 64];

	// get current date and time
	now = time( NULL );
	ptm = localtime( &now );

	va_start( ap, fmt );

#if 0
	if( opt_debug )
	{
		fprintf( stdout, "[%2d:%02d] ", ptm->tm_hour, ptm->tm_min );
		vfprintf( stdout, fmt, ap );
	}
#endif

	// allocate memory
	len = vsnprintf( NULL, 0, fmt, ap );
	len = len*5+10;
	buf1 = malloc( len );
	buf2 = malloc( len );

	// replace special charicter
	vsnprintf( buf1, allocated, fmt, ap );
	replace( buf1, buf2, '&', "&amp;" );
	replace( buf2, buf1, '<', "&lt;" );
	replace( buf1, buf2, '>', "&gt;" );

	// write to file
	sprintf( filename, "%s/%04d-%02d-%02d.log", opt_dirlog,
			ptm->tm_year+1900,
			ptm->tm_mon+1,
			ptm->tm_mday
			);
	out = fopen( filename, "a" );
	if( out != NULL )
	{
		fprintf( out, "<span class=\"irc-date\">[%2d:%02d]</span> <span class=\"%s\">",
				ptm->tm_hour, ptm->tm_min,
				class
		     );
		fputs( buf2, out );
		fputs( "</span><br />\n", out );

		fclose( out );
	}

	va_end( ap );

	free( buf1 );
	free( buf2 );
}

struct _event_callback **root_callback;

int register_event( event_callback_t *callback )
{
	event_callback_t *tmp;
	int length;

	if( root_callback == NULL )
		root_callback = calloc( sizeof(*root_callback), 1 );

	tmp = *root_callback;

	for( length=0; root_callback[length]; length++ )
	{
		if( !strcmp( root_callback[length]->eventname, callback->eventname ) )
		{
			// found same event list...
			// add it to the end.
			tmp = root_callback[length];

			while( tmp->next )
				tmp = tmp->next;
			tmp->next = callback;

			return 0;
		}
		length ++;
	}

	root_callback = realloc( root_callback, sizeof(*root_callback)*(length+2) );
	root_callback[length] = callback;
	root_callback[length+1] = NULL;

	return 0;
}

int unregister_event( event_callback_t *callback )
{
	event_callback_t *tmp;
	int a;

	for( a=0; root_callback[a]; a++ )
	{
		tmp = root_callback[a];
		if( tmp == callback )
		{
			if( tmp->next == NULL )
			{
				// remove the entire event name
				do
				{
					root_callback[a] = root_callback[a+1];
					a++;
				}
				while( root_callback[a] );
			}
			else
				// remove it just in the linked list
				root_callback[a] = tmp->next;

			return 0;
		}
		do
		{
			if( tmp->next == callback )
			{
				// found. remove it in the list
				tmp->next = callback->next;

				return 0;
			}

			tmp = tmp->next;
		}
		while( tmp->next );
	}

	return 0;
}

int execute_event( BN_PInfo I, const char *eventname, const char **args )
{
	event_callback_t *tmp;
	int a;

	// search the event list
	for( a=0; root_callback[a]; a++ )
	{
		if( !strcmp( root_callback[a]->eventname, eventname ) )
		{
			// found the event list
			tmp = root_callback[a];
			do
			{
				if( tmp->callback( I, tmp->arg, args ) == 1 )
					return 1;
				tmp = tmp->next;
			}
			while( tmp );

			return 0;
		}
	}

	return -1;
}

static void OnConnected( BN_PInfo I, const char hostname[] )
{
	logging( BLACK, "* connected to %s", hostname );

	BN_EnableFloodProtection( I, 100, 1000, 60 );
	BN_Register( I, opt_nick, opt_user, opt_realname );
}

static void OnDisconnected( BN_PInfo I, const char msg[] )
{
	logging( BLACK, "* disconnected.(%s)", msg );
}

static void OnRegistered( BN_PInfo I )
{
	logging( BLACK, "* registered." );
	BN_SendJoinMessage( I, opt_channel, NULL );
}

static void OnKill( BN_PInfo I, const char who[], const char whom[], const char msg[] )
{
	char nick1[64];
	char nick2[64];
	BN_ExtractNick( whom, nick1, 64 );
	BN_ExtractNick( who, nick2, 64 );

	logging( GREEN, "* %s has been killed by %s.(%s)",
			nick1, nick2, msg );
}

static void OnKick( BN_PInfo I, const char chan[], const char who[], const char whom[], const char msg[] )
{
	char nick1[64];
	char nick2[64];
	BN_ExtractNick( whom, nick1, 64 );
	BN_ExtractNick( who, nick2, 64 );

	logging( GREEN, "* %s has been kicked from %s by %s.(%s)",
			nick1, chan, nick2, msg );
}

static void OnTopic( BN_PInfo I, const char chan[], const char who[], const char msg[] )
{
	char nick[64];
	BN_ExtractNick( who, nick, 64 );

	logging( GREEN, "* %s changes topic to \'%s\'",
			nick, msg );
}

static void OnPrivateTalk( BN_PInfo I, const char who[], const char whom[], const char msg[] )
{
	const char *args[] = { who, whom, msg, NULL, };

	if( execute_event( I, "PrivateTalk", args ) != 1 )
	{
		char nick[64];
		BN_ExtractNick( who, nick, 64 );

		logging( BLACK, "<- *%s* %s",
				nick, msg );
	}
}

static void OnChannelTalk( BN_PInfo I, const char chan[], const char who[], const char msg[] )
{
	const char *args[] = { chan, who, msg, NULL, };

	if( execute_event( I, "ChannelTalk", args ) != 1 )
	{
		char nick[64];
		BN_ExtractNick( who, nick, 64 );

		logging( BLACK, "<%s> %s", nick, msg );
	}
}

static void OnNotice( BN_PInfo I, const char who[], const char whom[], const char msg[] )
{
	char nick[64];
	BN_ExtractNick( who, nick, 64 );

	logging( BROWN, "-%s- %s",
			nick, msg );
}

static void OnNick( BN_PInfo I, const char who[], const char msg[] )
{
	char nick[64];
	BN_ExtractNick( who, nick, 64 );

	logging( GREEN, "* %s is now known as %s",
			nick, msg );
}

static void OnJoin( BN_PInfo I, const char chan[], const char who[] )
{
	const char *args[] = { chan, who, NULL };

	if( execute_event( I, "Join", args ) != 1 )
	{
		char nick[64];
		char login[64];
		char host[64];

		BN_ExtractNick( who, nick, 64 );
		BN_ExtractUserName( who, login, 64 );
		BN_ExtractHost( who, host, 64 );

		logging( GREEN, "* %s (%s@%s) has joined in %s",
				nick, login, host, chan );
	}
}

static void OnQuit( BN_PInfo I, const char who[], const char msg[] )
{
	char nick[64];
	char login[64];
	char host[64];

	BN_ExtractNick( who, nick, 64 );
	BN_ExtractUserName( who, login, 64 );
	BN_ExtractHost( who, host, 64 );

	logging( NAVY, "* %s (%s@%s) Quit.(%s)",
			nick, login, host, msg );
}

static void OnPart( BN_PInfo I, const char chan[], const char who[], const char msg[] )
{
	char nick[64];
	char login[64];
	char host[64];

	BN_ExtractNick( who, nick, 64 );
	BN_ExtractUserName( who, login, 64 );
	BN_ExtractHost( who, host, 64 );

	logging( GREEN, "* %s (%s@%s) has left %s.(%s)",
			nick, login, host, chan, msg );
}

static void OnAction( BN_PInfo I, const char chan[], const char who[], const char msg[] )
{
	char nick[64];
	BN_ExtractNick( who, nick, 64 );

	logging( BRICK, "* %s %s", nick, msg );
}

void helpmsg( char *prgname )
{
	printf(

"%s [-d] [-h] [-D] [-l logdir] [-p port] [-c #channel] [-n nick] [-u user] [-r realname] <server>\n"
"  -d : make daemon\n"
"  -h : help message\n"
"  -D : run in debug mode\n"
"  -l \"log dir\"      : directory name to make log. default: current directory\n"
"  -a \"addon dir\"    : directory name to search add-on. default: current directory\n"
"  -p \"port number\"  : port number to connect to the server. default: 6667\n"
"  -c \"channel name\" : channel name to join. default: \"#irc\"\n"
"  -n \"nick name\"    : nick name to use. default: \"poors\"\n"
"  -u \"user name\"    : user name to use. default: \"username\"\n"
"  -r \"real name\"    : real name to use. default: \"realname\"\n"
"  \"server address\"  : server address to connect. default: \"localhost\"\n"

, prgname
);
}

int main( int argc, char **argv )
{
	BN_TInfo info;
	int ret;

	// get user arguments
	do
	{

		ret = getopt( argc, argv, "dhDp:c:n:u:r:l:a:" );
		switch( ret )
		{
			case -1:
				break;
			case 'h':
			default:
				helpmsg( argv[0] );
				exit( 0 );
				break;
			case 'd':
				opt_daemon = 1;
				break;
			case 'D':
				opt_debug ++;
				break;
			case 'p':
				opt_port = atoi( optarg );
				break;
			case 'c':
				opt_channel = optarg;
				break;
			case 'u':
				opt_user = optarg;
				break;
			case 'r':
				opt_realname = optarg;
				break;
			case 'n':
				opt_nick = optarg;
				break;
			case 'l':
				opt_dirlog = optarg;
				break;
			case 'a':
				opt_diraddon = optarg;
				break;
		}
	}
	while( ret >= 0 );
	if( optind < argc )
		opt_server = argv[optind];

	debug( "debug    %d\n", opt_debug );
	debug( "daemon   %d\n", opt_daemon );
	debug( "port     %d\n", opt_port );
	debug( "channel  \"%s\"\n", opt_channel );
	debug( "nick     \"%s\"\n", opt_nick );
	debug( "user     \"%s\"\n", opt_user );
	debug( "realname \"%s\"\n", opt_realname );
	debug( "server   \"%s\"\n", opt_server );

	// initialize botnet library
	memset( &info, 0, sizeof(info) );
	info.CB.OnConnected = OnConnected;
	info.CB.OnDisconnected = OnDisconnected;
	info.CB.OnRegistered = OnRegistered;
	info.CB.OnKick = OnKick;
	info.CB.OnKill = OnKill;
	info.CB.OnTopic = OnTopic;
	info.CB.OnPrivateTalk = OnPrivateTalk;
	info.CB.OnChannelTalk = OnChannelTalk;
	info.CB.OnNotice = OnNotice;
	info.CB.OnNick = OnNick;
	info.CB.OnJoin = OnJoin;
	info.CB.OnQuit = OnQuit;
	info.CB.OnPart = OnPart;
	info.CB.OnAction = OnAction;

	// for default actions
	extern int init_default_functions( void );
	init_default_functions();

	// connect to the server...
	while( BN_Connect( &info, opt_server, opt_port, 0 ) != true )
	{
		debug( "disconnected...\n" );
		sleep( 10 );
		debug( "retry...\n" );
	}

	return 0;
}

